           (*-------------------------------------------*
            |               November 2007               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory CSP_F_law_pipe_v4_1
imports CSP_F_law_aux CSP_F_op_pipe_v4_1 CSP_T_law_alpha_par
begin

(*****************************************************************

         1. mono & cong
         2. commutativity for Pipe
         3. associativity for Pipe

 *****************************************************************)

(* ============================================================== *
                          monotonicity
 * ============================================================== *)

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Pipe_mono:
  "[| left1 = left2 ;mid1 = mid2 ;right1 = right2 ;
      P1 <=T[M1,M2] Q1 ; 
      P2 <=T[M1,M2] Q2 |]
           ==> P1 <[left1,mid1,right1]> P2 <=T[M1,M2] Q1 <[left2,mid2,right2]> Q2"
apply (simp add: Pipe_def)
apply (simp add: cspT_decompo_Alpha_parallel cspT_decompo)
done

lemma cspT_Pipe_cong:
  "[| left1 = left2 ;mid1 = mid2 ;right1 = right2 ;
      P1 =T[M1,M2] Q1 ; 
      P2 =T[M1,M2] Q2 |]
           ==> P1 <[left1,mid1,right1]> P2 =T[M1,M2] Q1 <[left2,mid2,right2]> Q2"
apply (simp add: Pipe_def)
apply (simp add: cspT_decompo_Alpha_parallel cspT_decompo)
done

lemmas cspT_decompo_Pipe = cspT_Pipe_mono cspT_Pipe_cong

(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Pipe_mono:
  "[| left1 = left2 ;mid1 = mid2 ;right1 = right2 ;
      P1 <=F[M1,M2] Q1 ; 
      P2 <=F[M1,M2] Q2 |]
           ==> P1 <[left1,mid1,right1]> P2 <=F[M1,M2] Q1 <[left2,mid2,right2]> Q2"
apply (simp add: Pipe_def)
apply (simp add: cspF_decompo_Alpha_parallel cspF_decompo)
done

lemma cspF_Pipe_cong:
  "[| left1 = left2 ;mid1 = mid2 ;right1 = right2 ;
      P1 =F[M1,M2] Q1 ; 
      P2 =F[M1,M2] Q2 |]
           ==> P1 <[left1,mid1,right1]> P2 =F[M1,M2] Q1 <[left2,mid2,right2]> Q2"
apply (simp add: Pipe_def)
apply (simp add: cspF_decompo_Alpha_parallel cspF_decompo)
done

lemmas cspF_decompo_Pipe = cspF_Pipe_mono cspF_Pipe_cong

(* ============================================================== *
                       commutativity for Pipe
 * ============================================================== *)

(* ------------------- T ------------------- *)

lemma pipe_condition_prefix_sym_insert:
  "ALL left right s12 s1 s2.
   (inj left & inj right &
   pipe_condition_prefix s12 s1 s2 left right)
   --> pipe_condition_prefix s12 s2 s1 right left"
apply (simp add: pipe_condition_prefix_def)
apply (intro allI impI)
apply (elim conjE exE)
apply (drule_tac x="u'" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="t'" in exI, simp)
apply (rule_tac x="s'" in exI, simp)
apply (simp add: ren_tr_Renaming_channel_sym)
apply (simp add: Renaming_channel_commut)
done

lemma cspT_Pipe_commut_sub:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      P <[left,mid,right]> Q <=T[M,M] Q <[right,mid,left]> P"
apply (simp add: cspT_semantics)
apply (elim add_not_eq_symE)
apply (rule)
 apply (simp add: in_traces_Pipe)
 apply (auto)
 apply (rule_tac x="ta" in exI, simp)
 apply (rule_tac x="s" in exI, simp)
 apply (simp add: Un_commute)
 apply (simp add: ren_tr_Renaming_channel_sym)
 apply (simp add: Renaming_channel_commut)

 apply (insert pipe_condition_prefix_sym_insert)
  apply (drule_tac x="right" in spec)
  apply (drule_tac x="left" in spec)
  apply (drule_tac x="t" in spec)
  apply (drule_tac x="s" in spec)
  apply (drule_tac x="ta" in spec)
 apply (simp)
done

(*-----------------------------------*
 |         Commutativity (T)         |
 *-----------------------------------*)

lemma cspT_Pipe_commut:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      P <[left,mid,right]> Q =T[M,M] Q <[right,mid,left]> P"
apply (rule cspT_ref_eq)
apply (simp add: cspT_Pipe_commut_sub)
apply (elim add_not_eq_symE)
apply (simp add: cspT_Pipe_commut_sub)
done

(* ------------------- F ------------------- *)

lemma cspF_Pipe_commut_sub:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      P <[left,mid,right]> Q <=F[M,M] Q <[right,mid,left]> P"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Pipe_commut_sub)
apply (elim add_not_eq_symE)
apply (rule)
 apply (simp add: in_failures_Pipe)
 apply (elim conjE exE)
 apply (rule_tac x="t" in exI)
 apply (rule_tac x="sa" in exI)
 apply (rule_tac x="Y" in exI)
 apply (simp)
 apply (rule_tac x="Xa" in exI)
 apply (simp)
 apply (simp add: Un_commute)
 apply (simp add: ren_tr_Renaming_channel_sym)
 apply (simp add: Renaming_channel_commut)

 apply (insert pipe_condition_prefix_sym_insert)
  apply (drule_tac x="right" in spec)
  apply (drule_tac x="left" in spec)
  apply (drule_tac x="s" in spec)
  apply (drule_tac x="sa" in spec)
  apply (drule_tac x="t" in spec)
 apply (simp)
done

(*-----------------------------------*
 |         Commutativity (F)         |
 *-----------------------------------*)

lemma cspF_Pipe_commut:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      P <[left,mid,right]> Q =F[M,M] Q <[right,mid,left]> P"
apply (rule cspF_ref_eq)
apply (simp add: cspF_Pipe_commut_sub)
apply (elim add_not_eq_symE)
apply (simp add: cspF_Pipe_commut_sub)
done

(* ============================================================== *
                       associativity for Pipe
 * ============================================================== *)

(* -------------------------------------------- *
         lm for assoc (Tick or noTick)
 * -------------------------------------------- *)

lemma pipe_condition_prefix_assoc_noTick_or_Tick:
  "[| s123 rest-tr range left = s1 rest-tr range left ;
      s23 rest-tr range right = s123 rest-tr range right ;
       s23 rest-tr range left = s2 rest-tr range left ;
       s3 rest-tr range right = s23 rest-tr range right |]
    ==> noTick s1 & noTick s2 & noTick s3 & noTick s23 & noTick s123 |
        ~ noTick s1 & ~ noTick s2 & ~ noTick s3 & ~ noTick s23 & ~ noTick s123"
apply (case_tac "noTick s123")
apply (simp)
apply (insert rest_tr_noTick_left[of s123 "range left" s1 "range left"], simp)
apply (insert rest_tr_noTick_right[of s23 "range right" s123 "range right"], simp)
apply (insert rest_tr_noTick_left[of s23 "range left" s2 "range left"], simp)
apply (insert rest_tr_noTick_right[of s3 "range right" s123 "range right"], simp)
apply (simp)
apply (insert rest_tr_not_noTick_left[of s123 "range left" s1 "range left"], simp)
apply (insert rest_tr_not_noTick_right[of s23 "range right" s123 "range right"], simp)
apply (insert rest_tr_not_noTick_left[of s23 "range left" s2 "range left"], simp)
apply (insert rest_tr_not_noTick_right[of s3 "range right" s123 "range right"], simp)
done

(* -------------------------------------------- *
          pipe_condition (assoc) noTick
 * -------------------------------------------- *)

lemma pipe_condition_prefix_assoc_noTick:
  "ALL n s1 s2 s3 s23 s123.
     (inj left & inj right &
      (ALL x y. right x ~= left y)  &
      noTick s1 & noTick s2 & noTick s3 & noTick s23 & noTick s123 &
      lengtht s1 + lengtht s2 + lengtht s3 + lengtht s23 + lengtht s123 <= n &
      sett s1 <= insert Tick (Ev ` (range left Un range right)) &
      sett s2 <= insert Tick (Ev ` (range left Un range right)) &
      sett s3 <= insert Tick (Ev ` (range left Un range right)) &
      sett s23 <= insert Tick (Ev ` (range left Un range right)) &
      sett s123 <= insert Tick (Ev ` (range left Un range right)) &
        s123 rest-tr range left = s1 rest-tr range left &
          (s1 rest-tr range right) [[right<-->left]]* (s23 rest-tr range left) &
        s23 rest-tr range right = s123 rest-tr range right &
         s23 rest-tr range left = s2 rest-tr range left &
          (s2 rest-tr range right) [[right<-->left]]* (s3 rest-tr range left) &
         s3 rest-tr range right = s23 rest-tr range right &
      pipe_condition_prefix s123 s1 s23 left right &
      pipe_condition_prefix s23 s2 s3 left right)
--> (EX s12. noTick s12 &
             sett s12 <= insert Tick (Ev ` (range left Un range right)) &
             s12 rest-tr range left = s1 rest-tr range left &
             s2 rest-tr range right = s12 rest-tr range right &
             pipe_condition_prefix s12 s1 s2 left right &
             pipe_condition_prefix s123 s12 s3 left right)"

apply (rule)
apply (induct_tac n)
(* 0 *)
 apply (simp add: lengtht_zero)
 apply (intro allI impI)
 apply (simp add: pipe_condition_prefix_def)
 apply (rule_tac x="<>" in exI)
 apply (simp)

(* Suc *)
apply (intro allI impI)
apply (elim conjE exE)
apply (insert trace_nil_or_Tick_or_Ev)
apply (rotate_tac -1)
apply (frule_tac x="s1" in spec)
apply (frule_tac x="s2" in spec)
apply (frule_tac x="s3" in spec)
apply (frule_tac x="s23" in spec)
apply (drule_tac x="s123" in spec)
apply (elim add_not_eq_symE)

 (* s1 = <> *)
 apply (erule disjE)

  (* s2 = <> *)
  apply (erule disjE)

   (* s3 = <> *)
   apply (erule disjE)
   apply (simp)
   apply (rule_tac x="<>" in exI)
   apply (simp add: pipe_condition_prefix_def)
   apply (intro allI impI)
   apply (elim conjE exE)
   apply (rotate_tac 10)
   apply (drule_tac x="u'" in spec)
   apply (simp)
   apply (elim conjE exE)
   apply (simp)

   (* s3 = ... *)
   apply (rotate_tac -1)
   apply (erule disjE)
    apply (simp)
    apply (elim conjE exE)
    apply (simp add: image_iff)
    apply (elim conjE exE)
    apply (rotate_tac -4)
    apply (erule disjE)
     apply (elim conjE exE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)

    (* s23 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (erule disjE)
      apply (simp)
      apply (erule disjE)
      apply (simp)

      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)
       apply (elim conjE exE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp)

       (* s123 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -4)
        apply (erule disjE)
         apply (elim conjE exE)
         apply (simp)
         apply (elim conjE exE)
         apply (simp) 
         apply (drule_tac x="s1" in spec)
         apply (drule_tac x="s2" in spec)
         apply (drule_tac x="s" in spec)
         apply (drule_tac x="sa" in spec)
         apply (drule_tac x="sb" in spec)
         apply (simp add: pipe_condition_prefix_head)

  (* s2 = ... *)
 apply (rotate_tac -1)
 apply (erule disjE)
 apply (simp)
 apply (elim conjE exE)
 apply (simp add: image_iff)
 apply (elim conjE exE)
 apply (rotate_tac -4)
 apply (erule disjE)
  apply (elim conjE exE)
  apply (simp)
  apply (elim conjE exE)
  apply (simp)

   (* s3 = ... *)
   apply (erule disjE)
   apply (simp)
   apply (rotate_tac -1)
   apply (erule disjE)
   apply (simp)
   apply (elim conjE exE)
   apply (simp add: image_iff)
   apply (elim conjE exE)
   apply (rotate_tac -2)
   apply (erule disjE)
   (* s3 mid2 *)
    apply (elim conjE exE)
    apply (simp)

   (* s23 = <> *)
    apply (erule disjE)
    apply (elim conjE exE)
    apply (simp)

    (* s123 = <> *)
     apply (erule disjE)
     apply (simp)
     apply (drule_tac x="s1" in spec)
     apply (drule_tac x="s" in spec)
     apply (drule_tac x="sa" in spec)
     apply (drule_tac x="<>" in spec)
     apply (drule_tac x="<>" in spec)
     apply (simp add: pair_in_Renaming_channel)
     apply (simp add: pipe_condition_prefix_head)
     apply (elim conjE exE)
     apply (rule_tac x="<Ev (right xa)> ^^ s12" in exI)
     apply (simp add: image_iff)
     apply (simp add: pipe_condition_prefix_head)
     apply (fast)

    (* s123 = ... *)
     apply (erule disjE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp add: image_iff)
     apply (elim conjE exE)
     apply (rotate_tac -6)
     apply (erule disjE)
      apply (elim conjE exE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp)

   (* s23 = ... *)
    apply (rotate_tac -1)
    apply (erule disjE)
    apply (simp)
    apply (elim conjE exE)
    apply (simp)
    apply (simp add: image_iff)
    apply (elim conjE exE)
    apply (rotate_tac -4)
    apply (erule disjE)
     apply (elim conjE exE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)

    (* s123 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)
       apply (elim conjE exE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp)
       apply (drule_tac x="s1" in spec)
       apply (drule_tac x="s" in spec)
       apply (drule_tac x="sa" in spec)
       apply (drule_tac x="s23" in spec)
       apply (drule_tac x="s123" in spec)
       apply (simp add: pair_in_Renaming_channel)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rule_tac x="<Ev (right xa)> ^^ s12" in exI)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)

   (* s3 right *)
    apply (elim conjE exE)
    apply (simp)

   (* s23 = ... *)
    apply (erule disjE)
    apply (simp)
    apply (rotate_tac -1)
    apply (erule disjE)
    apply (simp)
    apply (elim conjE exE)
    apply (simp)
    apply (simp add: image_iff)
    apply (elim conjE exE)
    apply (rotate_tac -4)
    apply (erule disjE)
     apply (elim conjE exE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)

    (* s123 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)
       apply (elim conjE exE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp)
       apply (drule_tac x="s1" in spec)
       apply (drule_tac x="s2" in spec)
       apply (drule_tac x="sa" in spec)
       apply (drule_tac x="sb" in spec)
       apply (drule_tac x="sc" in spec)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)

 (* s1 = ... *)
 apply (rotate_tac -1)
 apply (erule disjE)
 apply (simp)
 apply (elim conjE exE)
 apply (simp add: image_iff)
 apply (elim conjE exE)
 apply (rotate_tac -2)
 apply (erule disjE)
  (* s1 left *)
  apply (elim conjE exE)
  apply (simp)

   (* s2 = <> *)
   apply (erule disjE)
   apply (simp)

    (* s3 = <> *)
    apply (erule disjE)
    apply (simp)

     (* s23 = <> *)
     apply (erule disjE)
     apply (simp)

      (* s123 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)
       apply (elim conjE exE)
       apply (simp)
       apply (drule_tac x="s" in spec)
       apply (drule_tac x="s2" in spec)
       apply (drule_tac x="s3" in spec)
       apply (drule_tac x="s23" in spec)
       apply (drule_tac x="sa" in spec)
       apply (simp add: pipe_condition_prefix_head)
       apply (elim conjE exE)
       apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
       apply (simp add: pipe_condition_prefix_head)

       apply (simp add: image_iff)

     (* s23 = ... *)
     apply (rotate_tac -1)
     apply (erule disjE)
     apply (simp)
     apply (erule disjE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp add: image_iff)
     apply (elim conjE exE)
     apply (rotate_tac -6)
     apply (erule disjE)
      apply (elim conjE exE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp)

    (* s3 = ... *)
    apply (rotate_tac -1)
    apply (erule disjE)
    apply (simp)
    apply (elim conjE exE)
    apply (simp add: image_iff)
    apply (elim conjE exE)
    apply (rotate_tac -4)
    apply (erule disjE)
     apply (elim conjE exE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)

      (* s23 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (rotate_tac -1)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)
       apply (elim conjE exE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp)

       (* s123 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -2)
        apply (erule disjE)

        (* s123 left *)
         apply (elim conjE exE)
         apply (simp)
         apply (drule_tac x="s" in spec)
         apply (drule_tac x="s2" in spec)
         apply (drule_tac x="s3" in spec)
         apply (drule_tac x="s23" in spec)
         apply (drule_tac x="sc" in spec)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)
         apply (elim conjE exE)
         apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)

        (* s123 right *)
         apply (elim conjE exE)
         apply (simp)
         apply (drule_tac x="s1" in spec)
         apply (drule_tac x="s2" in spec)
         apply (drule_tac x="sa" in spec)
         apply (drule_tac x="sb" in spec)
         apply (drule_tac x="sc" in spec)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)

   (* s2 = ... *)
   apply (rotate_tac -1)
   apply (erule disjE)
   apply (simp)
   apply (elim conjE exE)
   apply (simp add: image_iff)
   apply (elim conjE exE)
   apply (rotate_tac -2)
   apply (erule disjE)

    (* s2 left *)
    apply (elim conjE exE)
    apply (simp)

     (* s3 = <> *)
     apply (erule disjE)
     apply (simp)

      (* s23 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (rotate_tac -1)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)

       (* s23 left *)
       apply (elim conjE exE)
       apply (simp)
       apply (elim conjE exE)

        (* s123 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -4)
        apply (erule disjE)
   
         apply (elim conjE exE)
         apply (simp)
         apply (drule_tac x="s" in spec)
         apply (drule_tac x="s2" in spec)
         apply (drule_tac x="s3" in spec)
         apply (drule_tac x="s23" in spec)
         apply (drule_tac x="sc" in spec)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)
         apply (elim conjE exE)
         apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)

         apply (elim conjE exE)
         apply (simp)

       (* s23 right *)
       apply (elim conjE exE)
       apply (simp)

     (* s3 = ... *)
      apply (rotate_tac -1)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -2)
      apply (erule disjE) 

      (* s3 left *)
       apply (elim conjE exE)
       apply (simp)

       (* s23 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (rotate_tac -1)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -2)
        apply (erule disjE)

         (* s23 left *)
         apply (elim conjE exE)
         apply (simp)
         apply (elim conjE exE)

          (* s123 = ... *)
          apply (erule disjE)
          apply (simp)
          apply (erule disjE)
          apply (simp)
          apply (elim conjE exE)
          apply (simp add: image_iff)
          apply (elim conjE exE)
          apply (rotate_tac -2)
          apply (erule disjE)

           (* s123 left *)
           apply (elim conjE exE)
           apply (simp)
           apply (drule_tac x="s" in spec)
           apply (drule_tac x="s2" in spec)
           apply (drule_tac x="s3" in spec)
           apply (drule_tac x="s23" in spec)
           apply (drule_tac x="sd" in spec)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)
           apply (elim conjE exE)
           apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)

           (* s123 right *)
           apply (elim conjE exE)
           apply (simp add: prefix_pipe_condition_heads_false)

         (* s23 right *)
         apply (elim conjE exE)
         apply (simp)

          (* s123 = ... *)
          apply (erule disjE)
          apply (simp)
          apply (erule disjE)
          apply (simp)
          apply (elim conjE exE)
          apply (simp add: image_iff)
          apply (simp add: prefix_pipe_condition_heads_false)

      (* s3 right *)
       apply (elim conjE exE)
       apply (simp)

       (* s23 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (rotate_tac -1)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -2)
        apply (erule disjE)

         (* s23 left *)
         apply (elim conjE exE)
         apply (simp)
         apply (elim conjE exE)

          (* s123 = ... *)
          apply (erule disjE)
          apply (simp)
          apply (erule disjE)
          apply (simp)
          apply (elim conjE exE)
          apply (simp add: image_iff)
          apply (elim conjE exE)
          apply (rotate_tac -2)
          apply (erule disjE)

           (* s123 left *)
           apply (elim conjE exE)
           apply (simp)
           apply (drule_tac x="s" in spec)
           apply (drule_tac x="s2" in spec)
           apply (drule_tac x="s3" in spec)
           apply (drule_tac x="s23" in spec)
           apply (drule_tac x="sd" in spec)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)
           apply (elim conjE exE)
           apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)

           (* s123 right *)
           apply (elim conjE exE)
           apply (simp)
           apply (simp add: prefix_pipe_condition_heads_false)

         (* s23 right *)
         apply (elim conjE exE)
         apply (simp)

          (* s123 = ... *)
          apply (erule disjE)
          apply (simp)
          apply (erule disjE)
          apply (simp)
          apply (elim conjE exE)
          apply (simp add: image_iff)
          apply (elim conjE exE)
          apply (rotate_tac -2)
          apply (erule disjE)

           (* s123 left *)
           apply (elim conjE exE)
           apply (simp)
           apply (drule_tac x="s" in spec)
           apply (drule_tac x="s2" in spec)
           apply (drule_tac x="s3" in spec)
           apply (drule_tac x="s23" in spec)
           apply (drule_tac x="sd" in spec)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)
           apply (elim conjE exE)
           apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)

           (* s123 right *)
           apply (elim conjE exE)
           apply (simp)
           apply (drule_tac x="s1" in spec)
           apply (drule_tac x="s2" in spec)
           apply (drule_tac x="sb" in spec)
           apply (drule_tac x="sc" in spec)
           apply (drule_tac x="sd" in spec)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)

    (* s2 mid2 *)
    apply (elim conjE exE)
    apply (simp)

     (* s3 = ... *)
     apply (erule disjE)
     apply (simp)
     apply (rotate_tac -1)
     apply (erule disjE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)
     apply (simp add: image_iff)
     apply (elim conjE exE)
     apply (rotate_tac -2)
     apply (erule disjE)

      (* s3 left *)
      apply (elim conjE exE)
      apply (simp)

       apply (drule_tac x="s1" in spec)
       apply (drule_tac x="sa" in spec)
       apply (drule_tac x="sb" in spec)
       apply (drule_tac x="s23" in spec)
       apply (drule_tac x="s123" in spec)
       apply (simp add: pair_in_Renaming_channel)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rule_tac x="<Ev (right xb)> ^^ s12" in exI)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)

      (* s3 right *)
      apply (elim conjE exE)
      apply (simp)

      (* s23 = ... *)
       apply (erule disjE)
       apply (simp)
       apply (rotate_tac -1)
       apply (erule disjE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rotate_tac -2)
       apply (erule disjE)

        (* s23 left *)
        apply (elim conjE exE)
        apply (simp add: prefix_pipe_condition_heads_false)

        (* s23 right *)
        apply (elim conjE exE)
        apply (simp)

         (* s123 = ... *)
         apply (erule disjE)
         apply (simp)
         apply (erule disjE)
         apply (simp)
         apply (elim conjE exE)
         apply (simp add: image_iff)
         apply (elim conjE exE)
         apply (rotate_tac -2)
         apply (erule disjE)

          (* s123 left *)
          apply (elim conjE exE)
          apply (simp)
          apply (drule_tac x="s" in spec)
          apply (drule_tac x="s2" in spec)
          apply (drule_tac x="s3" in spec)
          apply (drule_tac x="s23" in spec)
          apply (drule_tac x="sd" in spec)
          apply (simp add: pipe_condition_prefix_head)
          apply (simp add: image_iff)
          apply (elim conjE exE)
          apply (rule_tac x="<Ev (left x)> ^^ s12" in exI)
          apply (simp add: pipe_condition_prefix_head)
          apply (simp add: image_iff)

          (* s123 right *)
          apply (elim conjE exE)
          apply (simp)
          apply (drule_tac x="s1" in spec)
          apply (drule_tac x="s2" in spec)
          apply (drule_tac x="sb" in spec)
          apply (drule_tac x="sc" in spec)
          apply (drule_tac x="sd" in spec)
          apply (simp add: pipe_condition_prefix_head)
          apply (simp add: image_iff)

  (* s1 right *)
  apply (elim conjE exE)
  apply (simp)

   (* s2 = ... *)
   apply (erule disjE)
   apply (simp)
   apply (rotate_tac -1)
   apply (erule disjE)
   apply (simp)
   apply (elim conjE exE)
   apply (simp add: image_iff)
   apply (elim conjE exE)
   apply (rotate_tac -2)
   apply (erule disjE)

    (* s2 left *)
    apply (elim conjE exE)
    apply (simp)

     (* s3 = <> *)
     apply (erule disjE)
     apply (simp)

      (* s23 = ... *)
      apply (erule disjE)
      apply (simp)
      apply (rotate_tac -1)
      apply (erule disjE)
      apply (simp)
      apply (elim conjE exE)
      apply (simp add: image_iff)
      apply (elim conjE exE)
      apply (rotate_tac -4)
      apply (erule disjE)

       (* s23 left *)
       apply (elim conjE exE)
       apply (simp)
       apply (drule_tac x="s" in spec)
       apply (drule_tac x="sa" in spec)
       apply (drule_tac x="s3" in spec)
       apply (drule_tac x="sb" in spec)
       apply (drule_tac x="s123" in spec)
       apply (simp add: pair_in_Renaming_channel)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)

     (* s3 = ... *)
     apply (rotate_tac -1)
     apply (erule disjE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp add: image_iff)
     apply (elim conjE exE)
     apply (rotate_tac -2)
     apply (erule disjE) 

      (* s3 mid2 *)
      apply (elim conjE exE)
      apply (simp)

       (* s23 = ... *)
       apply (erule disjE)
       apply (simp)
       apply (rotate_tac -1)
       apply (erule disjE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rotate_tac -2)
       apply (erule disjE)

        (* s23 left *)
        apply (elim conjE exE)
        apply (simp)
        apply (drule_tac x="s" in spec)
        apply (drule_tac x="sa" in spec)
        apply (drule_tac x="s3" in spec)
        apply (drule_tac x="sc" in spec)
        apply (drule_tac x="s123" in spec)
        apply (simp add: pair_in_Renaming_channel)
        apply (simp add: pipe_condition_prefix_head)
        apply (simp add: image_iff)

        (* s23 right *)
        apply (elim conjE exE)
        apply (simp add: prefix_pipe_condition_heads_false)

      (* s3 right *)
       apply (elim conjE exE)
       apply (simp)

       (* s23 = ... *)
        apply (erule disjE)
        apply (simp)
        apply (rotate_tac -1)
        apply (erule disjE)
        apply (simp)
        apply (elim conjE exE)
        apply (simp add: image_iff)
        apply (elim conjE exE)
        apply (rotate_tac -2)
        apply (erule disjE)

         (* s23 mid1 *)
         apply (elim conjE exE)
         apply (simp)
         apply (drule_tac x="s" in spec)
         apply (drule_tac x="sa" in spec)
         apply (drule_tac x="s3" in spec)
         apply (drule_tac x="sc" in spec)
         apply (drule_tac x="s123" in spec)
         apply (simp add: pair_in_Renaming_channel)
         apply (simp add: pipe_condition_prefix_head)
         apply (simp add: image_iff)

         (* s23 right *)
         apply (elim conjE exE)
         apply (simp)

          (* s123 = ... *)
          apply (erule disjE)
          apply (simp)
          apply (erule disjE)
          apply (simp)
          apply (elim conjE exE)
          apply (simp add: image_iff)
          apply (elim conjE exE)
          apply (rotate_tac -2)
          apply (erule disjE)

           (* s123 left *)
           apply (elim conjE exE)
           apply (simp add: prefix_pipe_condition_heads_false)

           (* s123 right *)
           apply (elim conjE exE)
           apply (simp)
           apply (drule_tac x="s1" in spec)
           apply (drule_tac x="s2" in spec)
           apply (drule_tac x="sb" in spec)
           apply (drule_tac x="sc" in spec)
           apply (drule_tac x="sd" in spec)
           apply (simp add: pair_in_Renaming_channel)
           apply (simp add: pipe_condition_prefix_head)
           apply (simp add: image_iff)

    (* s2 mid2 *)
    apply (elim conjE exE)
    apply (simp)

     (* s3 = ... *)
     apply (erule disjE)
     apply (simp)
     apply (rotate_tac -1)
     apply (erule disjE)
     apply (simp)
     apply (elim conjE exE)
     apply (simp)
     apply (simp add: image_iff)
     apply (elim conjE exE)
     apply (rotate_tac -2)
     apply (erule disjE)

      (* s3 mid2 *)
      apply (elim conjE exE)
      apply (simp)

       apply (drule_tac x="s1" in spec)
       apply (drule_tac x="sa" in spec)
       apply (drule_tac x="sb" in spec)
       apply (drule_tac x="s23" in spec)
       apply (drule_tac x="s123" in spec)
       apply (simp add: pair_in_Renaming_channel)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rule_tac x="<Ev (right xb)> ^^ s12" in exI)
       apply (simp add: pipe_condition_prefix_head)
       apply (simp add: image_iff)

      (* s3 right *)
      apply (elim conjE exE)
      apply (simp)

      (* s23 = ... *)
       apply (erule disjE)
       apply (simp)
       apply (rotate_tac -1)
       apply (erule disjE)
       apply (simp)
       apply (elim conjE exE)
       apply (simp add: image_iff)
       apply (elim conjE exE)
       apply (rotate_tac -2)
       apply (erule disjE)

        (* s23 mid1 *)
        apply (elim conjE exE)
        apply (simp add: prefix_pipe_condition_heads_false)

        (* s23 right *)
        apply (elim conjE exE)
        apply (simp)

         (* s123 = ... *)
         apply (erule disjE)
         apply (simp)
         apply (erule disjE)
         apply (simp)
         apply (elim conjE exE)
         apply (simp add: image_iff)
         apply (elim conjE exE)
         apply (rotate_tac -2)
         apply (erule disjE)

          (* s123 left *)
          apply (elim conjE exE)
          apply (simp add: prefix_pipe_condition_heads_false)

          (* s123 right *)
          apply (elim conjE exE)
          apply (simp)
          apply (drule_tac x="s1" in spec)
          apply (drule_tac x="s2" in spec)
          apply (drule_tac x="sb" in spec)
          apply (drule_tac x="sc" in spec)
          apply (drule_tac x="sd" in spec)
          apply (simp add: pipe_condition_prefix_head)
          apply (simp add: image_iff)
done

(* -------------------------------------------- *
            pipe_condition (assoc) Tick
 * -------------------------------------------- *)

lemma pipe_condition_prefix_assoc_Tick:
  "ALL s1 s2 s3 s23 s123.
     (inj left & inj right &
      (ALL x y. right x ~= left y)  &
      ~ noTick s1 & ~ noTick s2 & ~ noTick s3 & ~ noTick s23 & ~ noTick s123 &
      sett s1 <= insert Tick (Ev ` (range left Un range right)) &
      sett s2 <= insert Tick (Ev ` (range left Un range right)) &
      sett s3 <= insert Tick (Ev ` (range left Un range right)) &
      sett s23 <= insert Tick (Ev ` (range left Un range right)) &
      sett s123 <= insert Tick (Ev ` (range left Un range right)) &
        s123 rest-tr range left = s1 rest-tr range left &
          (s1 rest-tr range right) [[right<-->left]]* (s23 rest-tr range left) &
        s23 rest-tr range right = s123 rest-tr range right &
         s23 rest-tr range left = s2 rest-tr range left &
          (s2 rest-tr range right) [[right<-->left]]* (s3 rest-tr range left) &
         s3 rest-tr range right = s23 rest-tr range right &
      pipe_condition_prefix s123 s1 s23 left right &
      pipe_condition_prefix s23 s2 s3 left right)
--> (EX s12. sett s12 <= insert Tick (Ev ` (range left Un range right)) &
             s12 rest-tr range left = s1 rest-tr range left &
             s2 rest-tr range right = s12 rest-tr range right &
             pipe_condition_prefix s12 s1 s2 left right &
             pipe_condition_prefix s123 s12 s3 left right)"
apply (simp add: not_noTick_in_sett)
apply (intro allI impI)
apply (elim conjE exE)
apply (rename_tac s1 s2 s3 s23 s123 t1 t2 t3 t23 t123)
apply (simp)
apply (simp add: pipe_condition_prefix_simp_Tick)
apply (insert pipe_condition_prefix_assoc_noTick[of left right])
apply (drule_tac x=
 "lengtht t1 + lengtht t2 + lengtht t3 + lengtht t23 + lengtht t123" in spec)
apply (drule_tac x="t1" in spec)
apply (drule_tac x="t2" in spec)
apply (drule_tac x="t3" in spec)
apply (drule_tac x="t23" in spec)
apply (drule_tac x="t123" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="s12 ^^ <Tick>" in exI)
apply (simp add: pipe_condition_prefix_simp_Tick)
done

(* -------------------------------------------- *
               pipe_condition (assoc)
 * -------------------------------------------- *)

lemma pipe_condition_prefix_assoc:
  "ALL s1 s2 s3 s23 s123.
     (inj left & inj right &
      (ALL x y. right x ~= left y)  &
      sett s1 <= insert Tick (Ev ` (range left Un range right)) &
      sett s2 <= insert Tick (Ev ` (range left Un range right)) &
      sett s3 <= insert Tick (Ev ` (range left Un range right)) &
      sett s23 <= insert Tick (Ev ` (range left Un range right)) &
      sett s123 <= insert Tick (Ev ` (range left Un range right)) &
        s123 rest-tr range left = s1 rest-tr range left &
          (s1 rest-tr range right) [[right<-->left]]* (s23 rest-tr range left) &
        s23 rest-tr range right = s123 rest-tr range right &
         s23 rest-tr range left = s2 rest-tr range left &
          (s2 rest-tr range right) [[right<-->left]]* (s3 rest-tr range left) &
         s3 rest-tr range right = s23 rest-tr range right &
      pipe_condition_prefix s123 s1 s23 left right &
      pipe_condition_prefix s23 s2 s3 left right)
--> (EX s12. sett s12 <= insert Tick (Ev ` (range left Un range right)) &
             s12 rest-tr range left = s1 rest-tr range left &
             s2 rest-tr range right = s12 rest-tr range right &
             pipe_condition_prefix s12 s1 s2 left right &
             pipe_condition_prefix s123 s12 s3 left right)"
apply (intro allI impI)
apply (elim conjE exE)
apply (subgoal_tac 
  "(noTick s1 & noTick s2 & noTick s3 & noTick s23 & noTick s123 |
    ~ noTick s1 & ~ noTick s2 & ~ noTick s3 & ~ noTick s23 & ~ noTick s123)")
apply (erule disjE)
apply (insert pipe_condition_prefix_assoc_noTick[of left right])
apply (drule_tac x=
 "lengtht s1 + lengtht s2 + lengtht s3 + lengtht s23 + lengtht s123" in spec)
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)
apply (drule_tac x="s3" in spec)
apply (drule_tac x="s23" in spec)
apply (drule_tac x="s123" in spec)
apply (simp)
apply (elim conjE exE)
apply (rule_tac x="s12" in exI)
apply (simp)

apply (insert pipe_condition_prefix_assoc_Tick[of left right])
apply (drule_tac x="s1" in spec)
apply (drule_tac x="s2" in spec)
apply (drule_tac x="s3" in spec)
apply (drule_tac x="s23" in spec)
apply (drule_tac x="s123" in spec)
apply (simp)

apply (simp add: pipe_condition_prefix_assoc_noTick_or_Tick)
done

(* ======================== T ======================== *)

(*-----------------------------------*
 |         Associativity (sub)       |
 *-----------------------------------*)

lemma cspT_Pipe_assoc_sub:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      ((P <[left,mid,right]> Q) <[left,mid,right]> R) <=T[M,M]
      (P <[left,mid,right]> (Q <[left,mid,right]> R))"
apply (simp add: cspT_semantics)
apply (rule)

 apply (elim add_not_eq_symE)
 apply (simp add: in_traces_Pipe)
 apply (elim conjE exE)
 apply (rename_tac s123 s1 s23 s2 s3)
 apply (insert pipe_condition_prefix_assoc[of left right])
 apply (drule_tac x="s1" in spec)
 apply (drule_tac x="s2" in spec)
 apply (drule_tac x="s3" in spec)
 apply (drule_tac x="s23" in spec)
 apply (drule_tac x="s123" in spec)
 apply (simp)
 apply (elim conjE exE)

 apply (rule_tac x="s12" in exI)
 apply (rule)
 apply (rule_tac x="s1" in exI, simp)
 apply (rule_tac x="s2" in exI, simp)
 apply (rule_tac x="s3" in exI, simp)
done

(*-----------------------------------*
 |           Associativity           |
 *-----------------------------------*)
(*------------------*
 |      csp law     |
 *------------------*)

lemma cspT_Pipe_assoc:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      (P <[left,mid,right]> (Q <[left,mid,right]> R)) =T[M,M] 
      ((P <[left,mid,right]> Q) <[left,mid,right]> R)"
apply (elim add_not_eq_symE)
apply (rule cspT_ref_eq)

apply (rule cspT_rw_left)
apply (rule cspT_decompo_Pipe)
apply (simp)+
apply (rule cspT_reflex)
apply (rule cspT_Pipe_commut)
apply (simp)+
apply (rule cspT_rw_left)
apply (rule cspT_Pipe_commut)
apply (simp)+

apply (rule cspT_rw_right)
apply (rule cspT_decompo_Pipe)
apply (simp)+
apply (rule cspT_Pipe_commut)
apply (simp)+
apply (rule cspT_reflex)
apply (rule cspT_rw_right)
apply (rule cspT_Pipe_commut)
apply (simp)+
apply (rule cspT_Pipe_assoc_sub)
apply (simp)+

apply (simp add: cspT_Pipe_assoc_sub)
done

lemma cspT_Pipe_assoc_sym:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      ((P <[left,mid,right]> Q) <[left,mid,right]> R) =T[M,M] 
      (P <[left,mid,right]> (Q <[left,mid,right]> R))"
apply (rule cspT_sym)
apply (simp add: cspT_Pipe_assoc)
done

(* ======================== F ======================== *)

lemma cspF_Pipe_assoc_sub_set1:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y ;
      Ev ` range mid <= [[right<-->mid]]inv X1 Un [[left<-->mid]]inv Y ;
      Y Int Ev ` range left = X2 Int Ev ` range left |]
   ==> Ev ` range mid <= [[right<-->mid]]inv X1 Un [[left<-->mid]]inv X2"
apply (rule)
apply (erule subsetE)
apply (drule_tac x="x" in bspec, simp)
apply (simp)
apply (erule disjE, simp)
apply (simp add: image_iff)
apply (elim conjE exE)
apply (simp)
apply (simp add: ren_inv_def)
apply (elim conjE exE bexE)
apply (simp)
apply (elim add_not_eq_symE)
apply (simp add: pair_in_Renaming_channel)
apply (erule equalityE)
apply (erule subsetE)
apply (drule_tac x="Ev (left xa)" in bspec, simp)
apply (simp)
done

lemma cspF_Pipe_assoc_sub_set2:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y ;
      Ev ` range mid <= [[right<-->mid]]inv X2 Un [[left<-->mid]]inv X3 ;
      X23 Int Ev ` range left = X2 Int Ev ` range left;
      X23 Int Ev ` range right = X3 Int Ev ` range right |]
   ==> Ev ` range mid
           <= [[right<-->mid]]inv (X1 Int insert Tick (Ev ` range left)) Un
              [[right<-->mid]]inv (X2 Int insert Tick (Ev ` range right)) Un
              [[left<-->mid]]inv X3"
apply (elim add_not_eq_symE)
apply (rule)
apply (erule subsetE)
apply (drule_tac x="x" in bspec, simp)
apply (simp)
apply (rule disjI2)
apply (erule disjE)

 (* X2 *)
 apply (simp add: image_iff)
 apply (elim conjE exE)
 apply (simp)
 apply (simp add: ren_inv_def)
 apply (elim conjE exE bexE)
 apply (simp add: pair_in_Renaming_channel)

 (* X3 *)
 apply (simp add: image_iff)
done

(*-----------------------------------*
 |         Associativity (sub)       |
 *-----------------------------------*)

lemma cspF_Pipe_assoc_sub:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      ((P <[left,mid,right]> Q) <[left,mid,right]> R) <=F[M,M]
      (P <[left,mid,right]> (Q <[left,mid,right]> R))"
apply (simp add: cspF_cspT_semantics)
apply (simp add: cspT_Pipe_assoc_sub)
apply (rule)

 apply (elim add_not_eq_symE)
 apply (simp add: in_failures_Pipe)
 apply (elim conjE exE)
 apply (rename_tac s123 X123 s1 s23 X1 X23 s2 s3 X2 X3)
 apply (insert pipe_condition_prefix_assoc[of left right])
 apply (drule_tac x="s1" in spec)
 apply (drule_tac x="s2" in spec)
 apply (drule_tac x="s3" in spec)
 apply (drule_tac x="s23" in spec)
 apply (drule_tac x="s123" in spec)
 apply (simp)
 apply (elim conjE exE)

 apply (rule_tac x="s12" in exI)
 apply (rule_tac x="s3" in exI, simp)
 apply (rule_tac x="(X1 Int (insert Tick (Ev ` range left))) Un  
                    (X2 Int (insert Tick (Ev ` range right)))" in exI)
 apply (rule)
 apply (rule_tac x="s1" in exI, simp)
 apply (rule_tac x="s2" in exI, simp)
 apply (rule_tac x="X1" in exI, simp)
 apply (rule_tac x="X2" in exI, simp)

 apply (rule)
 apply (force)
 apply (rule)
 apply (force)
 apply (rule)
 apply (rule cspF_Pipe_assoc_sub_set1)
 apply (simp)+
 apply (force)

 apply (rule_tac x="X3" in exI, simp)
 apply (rule)
 apply (force)

 apply (rotate_tac -15)
 apply (drule sym)
 apply (simp)
 apply (simp add: cspF_Pipe_assoc_sub_set2)
 apply (auto)
done

(*-----------------------------------*
 |           Associativity           |
 *-----------------------------------*)
(*------------------*
 |      csp law     |
 *------------------*)

lemma cspF_Pipe_assoc:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      (P <[left,mid,right]> (Q <[left,mid,right]> R)) =F[M,M] 
      ((P <[left,mid,right]> Q) <[left,mid,right]> R)"
apply (elim add_not_eq_symE)
apply (rule cspF_ref_eq)

apply (rule cspF_rw_left)
apply (rule cspF_decompo_Pipe)
apply (simp)+
apply (rule cspF_reflex)
apply (rule cspF_Pipe_commut)
apply (simp)+
apply (rule cspF_rw_left)
apply (rule cspF_Pipe_commut)
apply (simp)+

apply (rule cspF_rw_right)
apply (rule cspF_decompo_Pipe)
apply (simp)+
apply (rule cspF_Pipe_commut)
apply (simp)+
apply (rule cspF_reflex)
apply (rule cspF_rw_right)
apply (rule cspF_Pipe_commut)
apply (simp)+
apply (rule cspF_Pipe_assoc_sub)
apply (simp)+

apply (simp add: cspF_Pipe_assoc_sub)
done

lemma cspF_Pipe_assoc_sym:
  "[| inj left; inj mid; inj right;
      ALL x y. right x ~= mid y ;
      ALL x y. left x ~= mid y ;
      ALL x y. right x ~= left y |] ==>
      ((P <[left,mid,right]> Q) <[left,mid,right]> R) =F[M,M] 
      (P <[left,mid,right]> (Q <[left,mid,right]> R))"
apply (rule cspF_sym)
apply (simp add: cspF_Pipe_assoc)
done

end
